{{
  Fast Walsh Hadamard Transform

  Jonathan "lonesock" Dummer
}}

VAR
  long  FWHT_control
  byte  cog
  
PUB start
  stop
  cog := cognew( @FWHT_pasm, @FWHT_control ) + 1
  return cog

PUB FWHT( vector_ptr, size_log2 )
  if size_log2 > 0
    FWHT_control := (size_log2 << 16) + vector_ptr
    repeat while FWHT_control

PUB sequency( index, bits )
  ' to go from Hadamard-ordered to sequency-ordered, take
  ' the bit reverse of the Gray-code of the index.
  return (((index >> 1) ^ index) >< bits)
  
PUB stop
  if cog
    cogstop( cog~ - 1 )

DAT
ORG 0
                        ' wait till we get the trigger
FWHT_pasm               rdlong  buf_len, par wz
              if_z      jmp     #FWHT_pasm

                        ' OK, reconstruct the address and size
                        mov     base_ptr, buf_len
                        shr     buf_len, #16

                        ' setup my outside loop
                        mov     stride, #1
                        
outer_loop              ' start the buffer pointers
                        mov     aptr, base_ptr
                        mov     bptr, base_ptr
                        add     bptr, stride                        
                        mov     sweep, #0

middle_loop             ' 1 sweep through                         
                        mov     bcount, stride

butterfly_loop          ' butterfly loop
                        rdlong  a, aptr
                        mov     tmp, a
                        rdlong  b, bptr
                        add     a, b
                        wrlong  a, aptr
                        add     aptr, #1
                        sub     tmp, b
                        wrlong  b, bptr
                        add     bptr, #1
                        djnz    bcount, #butterfly_loop
        
                        ' done with the middle loop (a single sweep)?
                        add     sweep, stride
                        add     sweep, stride 
                        cmp     sweep, buf_len wc,wz
              if_b      jmp     #middle_loop           

                        ' done with the outside loop?
                        shl     stride, #1
                        cmp     stride, buf_len wc,wz
              if_b      jmp     #outer_loop

                        ' DONE!
                        mov     tmp, #0
                        wrlong  tmp, par
                        jmp     #FWHT_pasm

' Scratch Variables              
buf_len       res       1
base_ptr      res       1
stride        res       1
sweep         res       1
bcount        res       1
aptr          res       1
bptr          res       1
a             res       1
b             res       1
tmp           res       1